/* scsicmd.c: generic SCSI command code
 * Copyright (C) 1996 Michael K. Johnson
 * 
 * I can be reached via email at johnsonm@redhat.com
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 * 
 */

#include <unistd.h>
#include <errno.h>
#include <stdio.h>

/* Get into kernel source tree with linux/../../  */
/* first need to create a dummy version of this structure that scsi.h wants.
 * Won't hurt us, just lets the compiler continue... */
struct request {
  char dummy;
};
#include <linux/../scsi/scsi.h>
#include <linux/../scsi/sg.h>

#include "scsicmd.h"
#include "debug.h"

static FILE *scsi_debug_out = NULL;

static int sfd = -1;
static int bufsize = BUFSIZE - (SCSI_OFF + 4);

/*
 * scsi_init initializes local variables.
 * 
 */

void scsi_init(int fd, int *size)
{
  /* sfd is the file descriptor to talk to the generic device on;
   * may be 0 if we are setting/querying bufsize and do not want
   * to change sfd... */
  if (fd) {
    sfd = fd;
  }
  if (size && *size > 0) {
    /* if a bufsize has been chosen, use it */
    bufsize = *size;
  }
  if (size) {
    /* if there is a variable to write to, report the current size,
     * which will be useful if it needs to be reset later */
    *size = bufsize;
  }
}

/* pass in file descriptor for debuggin */
void scsi_init_debug(FILE *f)
{
  scsi_debug_out = f;
}

/*
 * do_scsi_cmd sends a SCSI command.
 * Returns:
 *  -1 if unexpected result
 *  -2 if failed write
 *  -3 if failed read
 *  -4 if retryable failed write with bufsize reduced
 *  the response code otherwise
 *
 * obuf may be NULL if we don't care about the response...
 */

int do_scsi_cmd(unsigned in_size,      /* input data size */
		unsigned char *ibuf,   /* input buffer */
		unsigned out_size,     /* size of output buffer */
		unsigned char *obuf,   /* output buffer */
		int      expected      /* expected result, -1 for don't care */
		)
{
  int status = 0;
  struct sg_header *sgh;

  if (!obuf) out_size = 0;

  sgh = (struct sg_header *) ibuf;
  sgh->pack_len = SCSI_OFF + in_size;
  sgh->reply_len = SCSI_OFF + out_size;
  sgh->twelve_byte = 0; /* we assume this isn't necessary */

  if (scsi_debug_out)
    dump_send(ibuf + SCSI_OFF, in_size, scsi_debug_out);

  /* send command */
  status = write(sfd, ibuf, SCSI_OFF + in_size);
  if (status < 0 || status != SCSI_OFF + in_size ||
      ((expected >= 0) && sgh->result)) {
    if (errno == ENOMEM) {
      if (scsi_debug_out)
        dump_receive(obuf + SCSI_OFF, 0, sgh->result, scsi_debug_out);
      if (bufsize > 4096) {
	bufsize = 4096 - SCSI_OFF;
	return -4;
      }
    }
    fprintf(stderr, "write failed; CMD = 0x%x, result = 0x%x, errno = %d\n",
	    ibuf[SCSI_OFF], sgh->result, errno);
    perror("");
    return -2;
  }

  if (!obuf) obuf = ibuf;

  /* retrieve result */
  status = read(sfd, obuf, SCSI_OFF + out_size);
  if (status < 0 || status != SCSI_OFF + out_size ||
      ((sgh >= 0) && (expected >= 0) && (sgh->result != expected))) {
    fprintf(stderr, "read failed; CMD = 0x%x, result = 0x%x, errno = %d\n",
	    ibuf[SCSI_OFF], sgh->result);
    perror("");
#ifdef DEBUG
    fprintf(stderr, "sense code:\n"
	    "%x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x\n",
	    sgh->sense_buffer[0], sgh->sense_buffer[1],
	    sgh->sense_buffer[2], sgh->sense_buffer[3],
	    sgh->sense_buffer[4], sgh->sense_buffer[5],
	    sgh->sense_buffer[6], sgh->sense_buffer[7],
	    sgh->sense_buffer[8], sgh->sense_buffer[9],
	    sgh->sense_buffer[10], sgh->sense_buffer[11],
	    sgh->sense_buffer[12], sgh->sense_buffer[13],
	    sgh->sense_buffer[14], sgh->sense_buffer[15]);
#endif
    return -3;
  }
  if (!(status == SCSI_OFF + out_size)) {
    fprintf(stderr, "read failed to complete\n");
    return -3;
  }

  if (scsi_debug_out)
    dump_receive(obuf + SCSI_OFF, out_size, sgh->result, scsi_debug_out);

  return sgh->result;
}

