/* umax.c: scanner driver for UMAX scanners.
 * Copyright (C) 1995 Michael K. Johnson
 * Tested on my Vista-S6 (Envisions 6600S); should work on all
 * UMAX SCSI-2 scanners.  If it doesn't, please tell me and we'll
 * work it out, I hope...
 * 
 * I can be reached via email at johnsonm@redhat.com
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 * 
 */

#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <assert.h>
#include <signal.h>

/* Get into kernel source tree... */
/* first need to create a dummy version of this structure that scsi.h wants.
 * Won't hurt us, just lets the compiler continue... */
struct request {
  char dummy;
};
#include <linux/../scsi/scsi.h>
#include <linux/../scsi/sg.h>

#include "scsicmd.h" /* must come first; umax.h requires it */
#include "umax.h"
#include "debug.h"

#define min(a,b) (((a)<(b))?(a):(b))

/* All our lovely global variables... */
static unsigned char buffer[BUFSIZE];
#define bufcom (buffer + SCSI_OFF)
static unsigned int bufsize = -1;
static char *devicename = NULL;
static int verbose = 0;
static int colormode = 0; /* LINEART, GREYSCALE, or RGB */
#define LINEART 1
#define GREYSCALE 2
#define RGB 3
#define SCANNER_INCHES_WIDE 8.5
#define SCANNER_INCHES_TALL 11.7
static int row_len = -1;
static int width_in_pixels = -1;
static int height_in_pixels = -1;
static int phys_width = -1; /* 1200pt/inch */
static int phys_height = -1; /* 1200pt/inch */
static unsigned int max_value = 255;
static unsigned int x_resolution = 300;
static unsigned int y_resolution = 300;
static unsigned int upper_left_x = 0; /* 1200pt/inch */
static unsigned int upper_left_y = 0; /* 1200pt/inch */
static int brightness = -1;
static int threshold = -1;
static int ofd = STDOUT_FILENO; /* output file descriptor, stdout by default */


/* open_scanner tries to open the scanner.  It returns 0 on success,
 * and negative numbers on failure.
 * If no device name is specified on the command line, it hunts for
 * the first UMAX scanner on an sg device.
 * If a device name is specified on the command line, it trusts it
 * and does not test it as long as it can be opened.
 * If verbose output is enabled and it is searching for a scanner,
 * it tells the world where it found the scanner.
 */
int open_scanner(void) {
  int sfd;
  char filename[64]; /* only needs to hold "/dev/sg?"; this is overkill */
  char vendor[9];
  char product[0x10];
  char version[5];
  char c;

  if (devicename && strlen(devicename)) {
    /* user has specified name of device on command line, assume correct */
    sfd = open(devicename, O_RDWR);
    if (sfd == -1) { /* unsuccessful*/
      return -1;
    }
    /* success! */
    scsi_init(sfd, &bufsize);
    return 0;
  }
  for (c = 'a'; c <= 'h'; c++) {
    sprintf(filename, "/dev/sg%c", c);
    sfd = open(filename, O_RDWR);
    if (sfd != -1) {
      scsi_init(sfd, &bufsize);
      fullcmd(buffer, inquiry, inquiry_return_size, buffer, -1);
      get_inquiry_vendor(bufcom, vendor);
      get_inquiry_product(bufcom, product);
      get_inquiry_version(bufcom, version);
      if (strncmp("UMAX    ", vendor, 8)) {
        /* Not a UMAX product, try again. */
	close (sfd);
      } else {
	/* FIXME: test for scanner capabilities */
	if (verbose) {
	  /* in each case, first put a known space in, then replace
	   * the first space with a null byte to end the string. */
	  product[0xf] = ' ';
	  *(strchr(product, ' ')) = '\0';
	  version[4] = ' ';
	  *(strchr(version, ' ')) = '\0';
	  fprintf(stderr, "Found UMAX scanner %s version %s on device %s\n",
		  product, version, filename);
	}
	return 0;
      }
    }
  }
  /* if we reach here, we haven't found a UMAX scanner on any device */
  if (verbose) {
    fprintf(stderr, "No UMAX scanner found\n");
  }
  return -2;
}

/* wait_scanner should spin until TEST_INIT_READY returns GOOD
   returns 0 on success,
   returns -1 on error.  */

int wait_scanner(void) {
  int ret;
  int cnt = 0;

  do {
    scsicpy(buffer, test_unit_ready.cmd, test_unit_ready.size);
    ret = do_scsi_cmd(test_unit_ready.size, buffer, 0, NULL, -1);
    cnt++;
    if (cnt > 100) {
      fprintf(stderr, "testing unit; not ready\n");
      return -1;
    }
    if (ret == BUSY) {
      usleep(100000); /* .1 seconds */
    }
  } while (ret != GOOD);
  return 0;
}

#define WAIT_SCANNER ret = wait_scanner(); if (ret) return ret

/* grab_scanner should go through the following command sequence:
 * TEST UNIT READY
 *     CHECK CONDITION  \
 * REQUEST SENSE         > These should be handled automagically by
 *     UNIT ATTENTION   /  the kernel if they happen (powerup/reset)
 * TEST UNIT READY
 *     GOOD
 * RESERVE UNIT
 *     GOOD
 * 
 * It is then responsible for installing appropriate signal handlers
 * to call emergency_give_scanner() if user aborts.
 */
emergency_give_scanner(int i) {
  int ret;

  if (verbose) {
    fprintf(stderr, "emergency: trying to release scanner ...");
    fflush(stderr);
  }
  fullcmd(buffer, object_position, 0, NULL, GOOD);
  fullcmd(buffer, release_unit, 0, NULL, GOOD);
  if (verbose) {
    fprintf(stderr, " scanner released\n");
    fflush(stderr);
  }
  exit(1);
}

int grab_scanner(void) {
  int ret;
  int cnt = 0;

  do {
    scsicpy(buffer, test_unit_ready.cmd, test_unit_ready.size);
    ret = do_scsi_cmd(test_unit_ready.size, buffer, 0, NULL, -1);
    cnt++;
    if (cnt > 100) {
      fprintf(stderr, "testing unit; not ready\n");
      return -1;
    }
    if (ret == BUSY) {
      /* RS_sense_key_warmup is not implemented.
       * (we can simply assume that if it is busy we try back again
       * in a little bit...)
       * The test would be:
       * (RS_sense_key(((struct sg_header)buffer)->sense_buffer) & RS_sense_key_warmup)
       */
      usleep(100000); /* .1 seconds */
    }
  } while (ret != GOOD);
  fullcmd(buffer, reserve_unit, 0, NULL, GOOD);
  if (verbose) {
    fprintf(stderr, "unit tested ready after %d tries; unit reserved\n", cnt);
  }
  /* And if we abort? */
  signal(SIGINT, (__sighandler_t) emergency_give_scanner);
  signal(SIGTERM, (__sighandler_t) emergency_give_scanner);
  signal(SIGABRT, (__sighandler_t) emergency_give_scanner);
  signal(SIGILL, (__sighandler_t) emergency_give_scanner);
  signal(SIGSEGV, (__sighandler_t) emergency_give_scanner);
  return 0;
}

/* give_scanner should go through the following sequence:
 * OBJECT POSITION
 *     GOOD
 * RELEASE UNIT
 *     GOOD
 */

int give_scanner(void) {
  int ret;

  if (verbose) {
    fprintf(stderr, "trying to release scanner ...");
    fflush(stderr);
  }
  WAIT_SCANNER;
  fullcmd(buffer, object_position, 0, NULL, GOOD);
  WAIT_SCANNER;
  fullcmd(buffer, release_unit, 0, NULL, GOOD);
  if (verbose) {
    fprintf(stderr, " scanner released\n");
    fflush(stderr);
  }
  return 0;
}

/* set_window_param sets all the window parameters.  This
 * means building a fairly complicated SCSI command before
 * sending it...
 */
int set_window_param(void) {
  int num_dblocks = 1; /* number of window descriptor blocks, usually 1 or 3 */
  int cdblk = 0;
  int i;
  int ret;

  WAIT_SCANNER;
  if (colormode == RGB) num_dblocks = 3;
  scsicpy(buffer, set_window.cmd, set_window.size);
  scsicpy(WPDB_OFF(bufcom),
	  window_parameter_data_block.cmd, window_parameter_data_block.size);
  set_WPDB_wdbnum(WPDB_OFF(bufcom), num_dblocks);
  for (i = 1; i <= num_dblocks; i++) {
    scsicpy(WDB_OFF(buffer, i),
	    window_descriptor_block.cmd, window_descriptor_block.size);
    /* If there is only one WDB, then the WID is 0.
     * Otherwise, the WIDs are 1-3.  Yuck. */
    set_WD_wid(WDB_OFF(bufcom,i), (num_dblocks > 1)?i:0);
    set_WD_Xres(WDB_OFF(bufcom,i), x_resolution);
    set_WD_Yres(WDB_OFF(bufcom,i), y_resolution);
    set_WD_ULX(WDB_OFF(bufcom,i), upper_left_x);
    set_WD_ULY(WDB_OFF(bufcom,i), upper_left_y);
    set_WD_width(WDB_OFF(bufcom,i), phys_width);
    set_WD_length(WDB_OFF(bufcom,i), phys_height);
    set_WD_brightness(WDB_OFF(bufcom,i), brightness);
    set_WD_threshold(WDB_OFF(bufcom,i), threshold);
#ifdef LONG_WDB
    set_WD_pixel_count(WDB_OFF(bufcom, i), width_in_pixels);
    set_WD_line_count(WDB_OFF(bufcom, i), height_in_pixels);
#endif
    switch(colormode) {
     case LINEART:
      set_WD_composition(WDB_OFF(bufcom, i), WD_comp_lineart);
      set_WD_bitsperpixel(WDB_OFF(bufcom, i), WD_bits_1);
      set_WD_select_color(WDB_OFF(bufcom, i), WD_color_grey);
#ifdef LONG_WDB
      set_WD_calibration(WDB_OFF(bufcom, i), WD_calibration_lineart);
#endif
      break;
     case GREYSCALE:
      set_WD_composition(WDB_OFF(bufcom, i), WD_comp_grey);
      set_WD_bitsperpixel(WDB_OFF(bufcom, i), WD_bits_8);
      set_WD_select_color(WDB_OFF(bufcom, i), WD_color_grey);
#ifdef LONG_WDB
      set_WD_calibration(WDB_OFF(bufcom, i), WD_calibration_grey);
#endif
      break;
     case RGB:
      set_WD_composition(WDB_OFF(bufcom, i), WD_comp_rgb_full);
      set_WD_bitsperpixel(WDB_OFF(bufcom, i), WD_bits_8);
      switch (i) {
       /* first red, then green, then blue */
       case 1:
	set_WD_select_color(WDB_OFF(bufcom, i), WD_color_red);
	break;
       case 2:
	set_WD_select_color(WDB_OFF(bufcom, i), WD_color_green);
	break;
       case 3:
	set_WD_select_color(WDB_OFF(bufcom, i), WD_color_blue);
	break;
      }
#ifdef LONG_WDB
      set_WD_calibration(WDB_OFF(bufcom, i), WD_calibration_rgb);
#endif
      break;
    }
    if (verbose) {
      fprintf(stderr, "window descriptor block %d created\n", i);
    }
  }

  /* manipulate the command and window descriptor block(s) in the
   * buffer here */

  set_SW_xferlen(bufcom, (window_parameter_data_block.size +
			  (window_descriptor_block.size * num_dblocks)));
  docmd(set_window.size + window_parameter_data_block.size +
	(window_descriptor_block.size * num_dblocks), buffer, 0, NULL, GOOD);
  if (verbose) {
    fprintf(stderr, "window(s) set\n");
  }
  return 0;
}

int start_scan(void) {
  char typebuf[256];
  int widlistlen = 0; /* length of window id list */
  int i;
  int ret;

  WAIT_SCANNER;
  scsicpy(buffer, scan.cmd, scan.size);
#if 0
  /* The Envisions driver for the Vista S6 ALWAYS sets widlistlen
   * to 1 and makes the list element be a 0.  This works.
   * When I try to do that, the scanner aborts.  I don't know why.
   * At least for color and greyscale, setting widlistlen to 0
   * and having no list at all works.  <sigh>.  If anyone figures
   * this one out, please let me know.  If anyone wants to attack
   * the problem, I can send you traces.
   */
  switch(colormode) {
   case LINEART:
   case GREYSCALE:
    set_SC_wid(bufcom, 1, 0);
    widlistlen = 1;
    break;
   case RGB:
    for (i = 1; i <=3; i++) {
      set_SC_wid(bufcom, i, i);
    }
    widlistlen = 3;
    break;
  }
  /* manipulate the command in the buffer here to set
   * quality, use of ADF, and/or the preview bit */
  set_SC_xfer_length(bufcom, widlistlen);
#endif
  docmd(scan.size + widlistlen, buffer, 0, NULL, GOOD);
  WAIT_SCANNER;
  /* Now put out the appropriate header.
   * With the appropriate header, it looks like raw scanner data is
   * the appropriate format below, so no munging needs to take place.
   * If 30-bit format is added, it will have to be output in *ascii*
   * mode to use pnm utilities.  Yuck!
   */
  switch(colormode) {
   case LINEART:
    sprintf(typebuf, "P4\n# pbm created by umax\n%u %u\n",
	    width_in_pixels, height_in_pixels);
    break;
   case GREYSCALE:
    sprintf(typebuf, "P5\n# pgm created by umax\n%u %u\n%u\n",
	    width_in_pixels, height_in_pixels, max_value);
    break;
   case RGB:
    sprintf(typebuf, "P6\n# ppm created by umax\n%u %u\n%u\n",
	    width_in_pixels, height_in_pixels, max_value);
    break;
  }
  write(ofd, typebuf, strlen(typebuf));
  return 0;
}


int read_data(void) {
  unsigned int data_left;
  unsigned int data_to_read;
  unsigned int row_bufsize;
  unsigned int old_bufsize = bufsize;
  int ret;

  if (verbose) {
    fprintf(stderr, "Starting to READ data\n");
  }
  switch(colormode) {
   case LINEART:
    data_left = (width_in_pixels * height_in_pixels / 8);
    break;
   case GREYSCALE:
    /* one byte per pixel */
    data_left = width_in_pixels * height_in_pixels;
    break;
   case RGB:
    /* three bytes per pixel.  Note that high-end scanners
     * might have 6, if they are 30 bits deep total, since
     * they spread each pixel over two bytes.  No weird packing!
     * If you have one of these scanners, please work with me
     * to extend this program to deal with your extra bits.
     */
    data_left = width_in_pixels * height_in_pixels * 3;
    break;
  }

#if 0
  /* There is some idea that this might be required for some
   * scanners, so I'll leave it here for now.  If you find that
   * you have to use it, *tell me*, because if I don't hear
   * from anyone that it is necessary, it will be removed.
   */
  row_bufsize = (bufsize < row_len) ? bufsize - (bufsize % row_len)
                                    : bufsize;
#else
  row_bufsize = bufsize;
#endif
  if (verbose) {
    fprintf(stderr, "reading %u bytes in rows of %u\n",
	    data_left, row_bufsize);
  }

  do {
    /* if we want to, we can check to see if data_left agrees
     * with amount returned by GET DATA BUFFER STATUS, but why
     * bother? */
    WAIT_SCANNER;
#if 0 /* We aren't using this properly, and it's not necessary anyway.
       * Unless someone complains, this section will probably be
       * entirely removed in an upcoming release.
       */
    scsicpy(buffer, get_data_buffer_status.cmd, get_data_buffer_status.size);
    set_GDBS_len(bufcom, 1); /* one window; may need three on some scanners.
			      * Not sure; looking into this... */
    docmd(get_data_buffer_status.size, buffer, 14, buffer, GOOD);
    /* Do I care about what is returned?  Probably not, but I can
     * pretend for now... */
    if (verbose) {
      fprintf(stderr,
	      "Window ID: 0x%02x, ADB: %u, FDB: %u\n",
	      get_DBSD_winid(bufcom, 1),
	      get_DBSD_available_data_buffer(bufcom, 1),
	      get_DBSD_filled_data_buffer(bufcom, 1));
    }
#endif

    data_to_read = (data_left < row_bufsize) ? data_left : row_bufsize;

    WAIT_SCANNER;
    scsicpy(buffer, sread.cmd, sread.size);
    set_R_xfer_length(bufcom, data_to_read);
    ret = do_scsi_cmd(sread.size, buffer, data_to_read, buffer, GOOD);
    if (ret < 0) {
      if (ret == -4) {
	/* changed to smaller buffer size; try again */
        bufsize = -1; /* don't set bufsize back to original size! */
	scsi_init(0, &bufsize);
#if 0
	/* There is some idea that this might be required for some
	 * scanners, so I'll leave it here for now.  If you find that
         * you have to use it, *tell me*, because if I don't hear
	 * from anyone that it is necessary, it will be removed.
	 */
        row_bufsize = (bufsize < row_len) ? bufsize - (bufsize % row_len)
                                          : bufsize;
#else
	row_bufsize = bufsize;
#endif
	if (verbose) {
	  fprintf(stderr, "Retrying read: reading %u bytes in rows of %u\n",
		  data_left, row_bufsize);
	}
	continue;
      } else {
	/* some other non-recoverable error */
	return -1;
      }
    }
    /* output this chunk of data, which just happens to be
     * in pnm data format.  Yay! */
    write(ofd, bufcom, data_to_read);

    data_left -= data_to_read;
    if (verbose) {
      fprintf(stderr, "Buffer of %d bytes read; %d bytes to go\n",
              data_to_read, data_left);
    }
  } while (data_left);
  return 0;
}

/* Should call functions in this order:
 * open_scanner
 * grab_scanner
 * set_window_param
 * start_scan
 * read_data
 * give_scanner
 */

void usage(char * errormessage) {
  fprintf(stderr, "%s\n", errormessage);
  fprintf(stderr, "Usage: umax -lgr|-VH [options] [devicename]\n");
  fprintf(stderr, "l = lineart, g = greyscale, r = rgb\n"
		  "V = Version, H = Help\n"
		  "Options:\n"
	          "v = verbose, vv = very verbose (scsi command debugging)\n"
	          "p = preview quality (not yet implemented)\n"
	          "x = x resolution, y = y resolution, R = Resolution (both)\n"
	          "w = width in pixels, h = height in pixels\n"
	          "X = ULX, Y = ULY (1200 units per inch)\n"
	          "b = brightness (0), t = threshold (0x80)\n"
	          "[devicename] is optional; /dev/sg? are searched in order.\n"
	  );
  exit(1);
}


int main (int argc, char **argv) {
  FILE *debug_out;
  char c;

  if(argc < 2) {
    usage("too few arguments");
  }
  while ((c = getopt(argc, argv, "VHlgrvpx:y:R:w:h:X:Y:b:t:")) != EOF) {
    switch (c) {
    case 'V':
      usage("Umax driver version 0.4\n");
    case 'H':
      usage("Umax help:");
    case 'l':
      colormode = LINEART;
      break;
    case 'g':
      colormode = GREYSCALE;
      break;
    case 'r':
      colormode = RGB;
      break;
    case 'v':
      verbose++;
      break;
    case 'p':
      /* future preview mode */
      break;
    case 'x':
      x_resolution = atoi(optarg);
      break;
    case 'y':
      y_resolution = atoi(optarg);
      break;
    case 'R':
      x_resolution = y_resolution = atoi(optarg);
      break;
    case 'w':
      width_in_pixels = atoi(optarg);
      break;
    case 'h':
      height_in_pixels = atoi(optarg);
      break;
    case 'X':
      upper_left_x = atoi(optarg);
      break;
    case 'Y':
      upper_left_y = atoi(optarg);
      break;
    case 'b':
      brightness = atoi(optarg);
      break;
    case 't':
      threshold = atoi(optarg);
      break;
    case 0x3f:
      fprintf(stderr, "getopt complains about bad use of option :-(\n"
	              "Trying to continue anyway...\n");
      break;
    default:
      fprintf(stderr, "Unknown option '-%c' (ascii %02xh)\n", c, c);
      usage("bad option");
    }
  }
  if(optind < argc) {
    devicename = argv[optind];
  }

  if (colormode == 0) {
    usage ("Must choose lineart, greyscale, or rgb");
  }

  if (width_in_pixels < 0) {
    width_in_pixels = x_resolution * SCANNER_INCHES_WIDE;
  }
  if (height_in_pixels < 0) {
    height_in_pixels = y_resolution * SCANNER_INCHES_TALL;
  }

  /* bound width and height to stay within scanner, so that
   * -R n for n < 300 doesn't confuse the scanner */
  width_in_pixels = min(width_in_pixels,
			(x_resolution * SCANNER_INCHES_WIDE));
  height_in_pixels = min(height_in_pixels,
			 (y_resolution * SCANNER_INCHES_TALL));

  /* Now calculate the absolute width */
  phys_width = 1200 * width_in_pixels / x_resolution;
  phys_height = 1200 * height_in_pixels / y_resolution;

  if (verbose > 1) {
    debug_out = fopen("/tmp/umax.scsi.debug", "a");
    if (!debug_out) {
      /* open failed */
      usage ("Couldn't open /tmp/umax.scsi.debug");
    }
    scsi_init_debug(debug_out);
  }

  switch(colormode) {
   case LINEART:
    /* make sure line width divides evenly by 8... */
    width_in_pixels -= width_in_pixels % 8;
    phys_width = 1200 * width_in_pixels / x_resolution;
    row_len = (width_in_pixels / 8);
    if (brightness == -1)
      brightness = 0x20;
    if (threshold == -1)
      threshold = 0xe0;
    break;
   case GREYSCALE:
    /* one byte per pixel */
    row_len = width_in_pixels;
    if (brightness == -1)
      brightness = 0;
    if (threshold == -1)
      threshold = 0x80;
    break;
   case RGB:
    /* three bytes per pixel.  Note that high-end scanners
     * might have 6, if they are 30 bits deep total, since
     * they spread each pixel over two bytes.  No weird packing!
     * If you have one of these scanners, please work with me
     * to extend this program to deal with your extra bits.
     */
    row_len = width_in_pixels * 3;
    if (brightness == -1)
      brightness = 0;
    if (threshold == -1)
      threshold = 0x80;
    break;
  }

  if (verbose) {
    fprintf(stderr,
	    "Size- and position-related variables:\n"
	    "row_len = %u (measured in bytes)\n"
	    "width_in_pixels = %u\n"
	    "height_in_pixels = %u\n"
	    "phys_width = %u (1200pt/inch)\n"
	    "phys_height = %u (1200pt/inch)\n"
	    "x_resolution = %u (pixels/inch)\n"
	    "y_resolution = %u (pixels/inch)\n"
	    "upper_left_x = %u (offset 1200pt/inch)\n"
	    "upper_left_y = %u (offset 1200pt/inch)\n"
	    "brightness = %u, threshold = %u\n"
	    "\n",
	    row_len, width_in_pixels, height_in_pixels,
	    phys_width, phys_height,
	    x_resolution, y_resolution,
	    upper_left_x, upper_left_y,
	    brightness, threshold
	    );
  }

#define try_to(command, error) if(command) { fprintf(stderr, error); exit(1); }
  try_to (open_scanner(), "Could not open scanner\n");
  try_to (grab_scanner(), "Could not allocate scanner\n");
  try_to (set_window_param(), "Could not initialize scanner\n");
  try_to (start_scan(), "Could not start scan\n");
  try_to (read_data(), "Failed while attempting to read data\n");
  if (give_scanner()) {
    /* does not need to abort; got the scan, but bus or scanner is
     * probably hung... */
    fprintf (stderr, "Failed to deallocate scanner\n");
  }

  exit (0);
}
